<?php

namespace Drupal\mbp_convert\Form;

use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a confirmation form before clearing out the examples.
 */
class ParagraphConvertForm extends FormBase {

  const TARGET_PARAGRAPH_TYPES = [
    'mbp_container',
    'mbp_columns',
    'mbp_columns_2',
    'mbp_columns_3',
  ];

  /**
   * The entity_field.manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity_display.repository service.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * The layout plugin manager service.
   *
   * @var \Drupal\Core\Layout\LayoutPluginManagerInterface
   */
  protected $layoutPluginManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);

    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityDisplayRepository = $container->get('entity_display.repository');
    $instance->layoutPluginManager = $container->get('plugin.manager.core.layout');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mbp_convert_paragraph_convert';
  }

  /**
   * Get paragraph map for fields using the layout paragraphs widget.
   *
   * @return array
   *   Returns a map of paragraph fields using layout paragraph widgets.
   */
  protected function getLayoutParagraphFieldMap() : array {
    $map = [];

    foreach ($this->entityFieldManager->getFieldMapByFieldType('entity_reference_revisions') as $entity_type => $fields) {
      foreach ($fields as $field_name => $field_info) {
        $sd = $this->entityFieldManager->getFieldStorageDefinitions($entity_type);

        if (
          ($sd[$field_name] ?? NULL) &&
          $sd[$field_name]->getSetting('target_type') === 'paragraph'
        ) {
          foreach ($field_info['bundles'] as $bundle) {
            /** @var \Drupal\Core\Entity\Display\EntityViewDisplayInterface $form_mode */
            $form_mode = $this->entityTypeManager
              ->getStorage('entity_form_display')
              ->load("$entity_type.$bundle.default");
            $component = $form_mode->getComponent($field_name);

            if ($component['type'] === 'layout_paragraphs') {
              $map[$entity_type][$field_name][$bundle] = $bundle;
            }
          }
        }
      }
    }

    return $map;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['actions'] = ['#type' => 'actions'];

    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => 'Convert',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $layouts_map = $this->getLayoutParagraphFieldMap();

    if (!empty($layouts_map)) {
      $batch = [
        'operations' => [],
        'title' => $this->t('Converting Paragraphs'),
      ];

      foreach ($layouts_map as $entity_type => $fields) {
        foreach ($fields as $field_name => $bundles) {
          $batch['operations'][] = [
            [$this, 'layoutParagraphsBatch'],
            [
              $entity_type,
              $field_name,
              $bundles,
            ],
          ];
        }
      }

      if (!empty($batch['operations'])) {
        batch_set($batch);
      }
    }
  }

  /**
   * Batch operation.
   */
  public function layoutParagraphsBatch($entity_type, $field_name, $bundles, &$context) {
    $context['finished'] = 1;

    $ps = $this->entityTypeManager->getStorage('paragraph');
    $prids = $ps->getQuery()->allRevisions()
      ->condition('type', static::TARGET_PARAGRAPH_TYPES, 'IN')
      ->condition('parent_type', $entity_type)
      ->condition('parent_field_name', $field_name)
      ->execute();

    if (!empty($prids)) {
      $storage = $this->entityTypeManager->getStorage($entity_type);
      $eids = $storage->getQuery()->currentRevision()
        ->condition('type', $bundles, 'IN')
        ->condition("$field_name.target_id", array_values($prids), 'IN')
        ->execute();

      if ($context['sandbox']['total'] ?? NULL) {
        // Initialize the progress and immediately quit so you can see the 0%.
        $context['sandbox']['total'] = $context['sandbox']['total'] ?? count($eids);
        $context['sandbox']['left'] = count($eids);

        if (count($eids) > 1) {
          $context['finished'] = ($context['sandbox']['total'] - $context['sandbox']['left']) / $context['sandbox']['total'];
        }
        else {
          $context['finished'] = 1;
        }

        return;
      }

      $eids = array_slice($eids, 0, 5);

      if (
        !empty($eids) &&
        !empty($entities = $storage->loadMultiple($eids))
      ) {
        /** @var \Drupal\Core\Entity\FieldableEntityInterface[] $entities */

        foreach ($entities as $entity) {
          $new_paragraphs = [];
          $old_paragraphs = [];

          foreach ($entity->get($field_name) as $delta => $field_item) {
            /** @var \Drupal\paragraphs\ParagraphInterface $paragraph */
            $paragraph = $field_item->entity ?? NULL;

            if ($paragraph) {
              $layout = NULL;
              switch ($paragraph->bundle()) {
                case 'mbp_container':
                  $old_paragraphs[] = $old_paragraph = $paragraph;

                  $new_paragraphs[] = $paragraph = $ps->create([
                    'type' => 'mbp_layout',
                    'status' => 1,
                    'behavior_settings' => serialize([
                      'layout_paragraphs' => [
                        'region' => '',
                        'parent_uuid' => '',
                        'layout' => 'monarch_bootstrap_container',
                        'config' => [
                          'label' => '',
                          'container' => $paragraph->mbp_container->value,
                          'bg_color' => $paragraph->mbp_background_color->value,
                        ],
                      ],
                    ]),
                  ]);

                  foreach ($old_paragraph->mbp_elements->referencedEntities() as $child_paragraph) {
                    /** @var \Drupal\paragraphs\ParagraphInterface $child_paragraph */
                    $settings = $child_paragraph->getAllBehaviorSettings() ?: [];

                    $settings['layout_paragraphs'] = [
                      'region' => 'content',
                      'parent_uuid' => $paragraph->uuid(),
                      'layout' => '',
                    ] + ($settings['layout_paragraphs'] ?? []) + [
                      'config' => [],
                    ];

                    $child_paragraph->setAllBehaviorSettings($settings);
                    $child_paragraph->setNewRevision(TRUE);
                    $child_paragraph->save();
                    $new_paragraphs[] = $child_paragraph;
                  }

                  break;

                case 'mbp_columns_2':
                case 'mbp_columns_3':
                case 'mbp_columns':
                  $old_paragraphs[] = $old_paragraph = $paragraph;
                  $layout_option = NULL;

                  if ($old_paragraph->bundle() === 'mbp_columns_2') {
                    $layout = 'monarch_bootstrap_columns_2';
                    $source_field = 'mbp_elements_2';
                    $layout_option = [
                      'mbp-layout--50-50' => 'layout--50-50',
                      'mbp-layout--75-25' => 'layout--75-25',
                      'mbp-layout--66-33' => 'layout--66-33',
                      'mbp-layout--80-20' => 'layout--80-20',
                      'mbp-layout--25-75' => 'layout--25-75',
                      'mbp-layout--33-66' => 'layout--33-66',
                      'mbp-layout--20-80' => 'layout--20-80',
                    ][$old_paragraph->mbp_layout_2->value ?? NULL] ?? 'layout--50-50';
                  }
                  elseif ($old_paragraph->bundle() === 'mbp_columns_3') {
                    $layout = 'monarch_bootstrap_columns_3';
                    $source_field = 'mbp_elements_3';
                    $layout_option = [
                      'mbp-layout--33-33-33' => 'layout--33-33-33',
                      'mbp-layout--25-50-25' => 'layout--25-50-25',
                      'mbp-layout--50-25-25' => 'layout--50-25-25',
                      'mbp-layout--25-25-50' => 'layout--25-25-50',
                      'mbp-layout--16-66-16' => 'layout--16-66-16',
                      'mbp-layout--66-16-16' => 'layout--66-16-16',
                      'mbp-layout--16-16-66' => 'layout--16-16-66',
                    ][$old_paragraph->mbp_layout_3->value ?? NULL] ?? 'layout--33-33-33';
                  }
                  else {
                    $layout = 'monarch_bootstrap_columns';
                    $source_field = 'mbp_elements';
                  }

                  if ($layout_definition = ($this->layoutPluginManager->getSortedDefinitions()[$layout] ?? NULL)) {
                    $columns = array_keys($layout_definition->getRegions() ?: []);
                  }

                  if (empty($columns)) {
                    throw new \Exception('No regions found!');
                  }

                  $behavior_settings = $old_paragraph->getAllBehaviorSettings() ?: [];
                  $behavior_settings['layout_paragraphs']['layout'] = $layout;

                  if ($layout === 'monarch_bootstrap_container') {
                    $behavior_settings['layout_paragraphs']['config']['container'] = 'container-fluid';
                  }

                  if (!empty($old_paragraph->mbp_breakpoint->value ?? NULL)) {
                    $behavior_settings['layout_paragraphs']['config']['breakpoint'] = $old_paragraph->mbp_breakpoint->value;
                  }

                  if (!empty($layout_option)) {
                    $behavior_settings['layout_paragraphs']['config']['layout'] = $layout_option;
                  }

                  $new_paragraphs[] = $paragraph = $ps->create([
                    'type' => 'mbp_layout',
                    'status' => 1,
                    'behavior_settings' => serialize($behavior_settings),
                  ]);

                  reset($columns);

                  foreach ($old_paragraph->get($source_field)->referencedEntities() as $child_paragraph) {
                    /** @var \Drupal\paragraphs\ParagraphInterface $child_paragraph */
                    $settings = $child_paragraph->getAllBehaviorSettings() ?: [];

                    $settings['layout_paragraphs'] = [
                      'region' => current($columns),
                      'parent_uuid' => $paragraph->uuid(),
                      'layout' => '',
                    ] + ($settings['layout_paragraphs'] ?? []) + [
                      'config' => [],
                    ];

                    if (!next($columns)) {
                      reset($columns);
                    }

                    $child_paragraph->setAllBehaviorSettings($settings);
                    $child_paragraph->setNewRevision(TRUE);
                    $child_paragraph->save();
                    $new_paragraphs[] = $child_paragraph;
                  }

                  break;

                default:
                  $new_paragraphs[] = $paragraph;
                  break;
              }
            }
          }

          if (!empty($old_paragraphs)) {
            // Save new paragraphs here.
            foreach ($new_paragraphs as $index => $new_paragraph) {
              /** @var \Drupal\paragraphs\ParagraphInterface $new_paragraph */
              if (empty($new_paragraph->id())) {
                $new_paragraph->save();
              }

              $new_paragraphs[$index] = [
                'target_id' => $new_paragraph->id(),
                'target_revision_id' => $new_paragraph->revision_id->value,
              ];
            }

            foreach ($old_paragraphs as $index => $old_paragraph) {
              /** @var \Drupal\paragraphs\ParagraphInterface $old_paragraph */
              $old_paragraph->parent_type->value = NULL;
              $old_paragraph->parent_id->value = NULL;
              $old_paragraph->parent_field_name->value = NULL;
              $old_paragraph->setNewRevision(TRUE);
              $old_paragraph->save();
            }

            $entity->get($field_name)->setValue($new_paragraphs);

            if ($entity instanceof RevisionableInterface) {
              $entity->setNewRevision(TRUE);
            }

            $entity->save();
          }

          $context['sandbox']['left'] = max($context['sandbox']['left'] - 1, 0);
          $context['finished'] = $context['sandbox']['total'] ? ($context['sandbox']['total'] - $context['sandbox']['left']) / $context['sandbox']['total'] : 1;
        }
      }
    }
  }

}
